// main.cpp : Defines the entry point for the console application.
//

#include "config.h"
#include <commdlg.h>
#include <conio.h>
#include <iostream>

#include "cximage/ximage.h"
#include "xbrowseforfolder.h"

// function prototypes
bool CalculateDescriptors(const char *cat1dir, const char *cat2dir, const char *cat3dir);
bool CategorizeDescriptors(const char *cat1dir, const char *cat2dir, const char *cat3dir);
bool DetermineWinterSport(const char *cat1dir, const char *cat2dir, const char *cat3dir);

// names of categories
#define CATEGORY_1	"cat1.crowd"
#define CATEGORY_2	"cat2.skijump"
#define CATEGORY_3	"cat3.snowboarding"

// number of bins to use for each color
// Note: number of bins must be a whole fraction of 256. if we would
//       use 256 bins for each color then a single histogram would be
//       64MB, so we should choose a more sensible number
#define BIN_COUNT	32


int main(int argc, char **argv)
{
	char dbdir[MAX_PATH];
	// ask the user for the image database
	if (XBrowseForFolder(NULL, "Please select image database folder", NULL, dbdir, sizeof(dbdir)) == FALSE)
		return 0;
	if (strlen(dbdir) == 0)
		return 0;
	// in the image directory, the images are stored in categorized
	// folders. store the descriptors we calculate in the same
	// folder as the image
	char cat1dir[MAX_PATH], cat2dir[MAX_PATH], cat3dir[MAX_PATH];
	SAFE_SPRINTF(cat1dir, sizeof(cat1dir), "%s\\%s\\", dbdir, CATEGORY_1);
	SAFE_SPRINTF(cat2dir, sizeof(cat2dir), "%s\\%s\\", dbdir, CATEGORY_2);
	SAFE_SPRINTF(cat3dir, sizeof(cat3dir), "%s\\%s\\", dbdir, CATEGORY_3);
	// ask the user which option to use
	while (true)
	{
		system("cls");
		cout << "***************************" << endl;
		cout << "*  Winter Olympic Imagery *" << endl;
		cout << "***************************" << endl;
		cout << endl;
		cout << "1. calculate descriptors" << endl;
		cout << "2. categorize descriptors" << endl;
		cout << "3. determine winter sport" << endl;
		cout << endl;
		cout << "Please select option, or type 'q' to quit: ";
		char c = _getch();
		cout << c << endl;
		fflush(stdin);
		// start the chosen option
		switch (c)
		{
		case 'q':
			return 0;
		case '1':
			if (!CalculateDescriptors(cat1dir, cat2dir, cat3dir))
			{
				cout << "could not calculate descriptors" << endl;
				return 1;
			}
			break;
		case '2':
			if (!CategorizeDescriptors(cat1dir, cat2dir, cat3dir))
			{
				cout << "could not categorize descriptors" << endl;
				return 1;
			}
			break;
		case '3':
			if (!DetermineWinterSport(cat1dir, cat2dir, cat3dir))
			{
				cout << "could not determine winter sport" << endl;
				return 1;
			}
			break;
		default:
			continue;
		}
	}
	return 0;
}

// histogram should be a preallocated array of size BIN_COUNT*BIN_COUNT*BIN_COUNT elements and will
// be filled with the color histogram of the image where path points at
bool CalculateDescriptor(const char *path, float *histogram)
{
	// load the image
	CxImage image(path, CXIMAGE_FORMAT_JPG);
	if (!image.IsValid())
		return false;
	// clear histogram
	memset(histogram, 0, BIN_COUNT*BIN_COUNT*BIN_COUNT * sizeof(float));
	// walk through the pixels to fill the histogram
	int width = (int)image.GetWidth();
	int height = (int)image.GetHeight();
	int bin_r, bin_g, bin_b;
	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			// Note: CxImage library starts counting at lower-left corner of the image,
			//       which is seen as the top of the image. however, usually images
			//       start counting from the top-left corner of the image. thus if you
			//       want to get pixel(2, 2) from the top-left you would have to ask
			//       for pixel (2, height - 2 - 1) from CxImage. although in this
			//       situation we don't care which pixel is where, we only care about
			//       its color.
			RGBQUAD rgb = image.BlindGetPixelColor(x, y, false);
			// determine the bin this color falls in
			bin_r = rgb.rgbRed   / (256 / BIN_COUNT);
			bin_g = rgb.rgbGreen / (256 / BIN_COUNT);
			bin_b = rgb.rgbBlue  / (256 / BIN_COUNT);
			histogram[bin_r*BIN_COUNT*BIN_COUNT + bin_g*BIN_COUNT + bin_b]++;
		}
	}
	// normalize the histogram so that all together the values will add up
	// to one. since there are width * height pixels, we divide each value
	// by this amount
	for (int i = 0; i < BIN_COUNT*BIN_COUNT*BIN_COUNT; i++)
		histogram[i] /= width * height;
	return true;
}

bool CalculateDescriptors(const char *cat1dir, const char *cat2dir, const char *cat3dir)
{
	// the histogram that we reuse for each image
	float *histogram = new float[BIN_COUNT*BIN_COUNT*BIN_COUNT];
	// walk through all images
	// Note: each of the three categories has 50 images
	char path[MAX_PATH];
	const char *catdir, *catname;
	FILE *file = NULL;
	for (int c = 1; c <= 3; c++)
	{
		// determine the category directory to use
		switch (c)
		{
		case 1: catdir = cat1dir; catname = CATEGORY_1; break;
		case 2: catdir = cat2dir; catname = CATEGORY_2; break;
		case 3: catdir = cat3dir; catname = CATEGORY_3; break;
		}
		cout << "processing category " << catname << endl;
		// process the images in the directory
		for (int i = 1; i <= 50; i++)
		{
			cout << "processing image " << i << endl;
			// calculate the histogram descriptor
			SAFE_SPRINTF(path, sizeof(path), "%s%i.jpg", catdir, i);
			if (!CalculateDescriptor(path, histogram))
				goto failure;
			// save the descriptor to disk
			SAFE_SPRINTF(path, sizeof(path), "%s%i.dat", catdir, i);
			if ((file = fopen(path, "wb")) == NULL)
				goto failure;
			if (fwrite(histogram, sizeof(float), BIN_COUNT*BIN_COUNT*BIN_COUNT, file) != BIN_COUNT*BIN_COUNT*BIN_COUNT)
				goto failure;
			SAFE_CLOSEFILE(file);
		}
	}
	// release resources
	SAFE_DELETE_ARRAY(histogram);
	return true;

failure:
	SAFE_CLOSEFILE(file);
	SAFE_DELETE_ARRAY(histogram);
	return false;
}

bool CategorizeDescriptors(const char *cat1dir, const char *cat2dir, const char *cat3dir)
{
	// analyze the descriptors per category to determine the
	// characteristics of that category
	float *histogram = new float[BIN_COUNT*BIN_COUNT*BIN_COUNT];
	float *average = new float[BIN_COUNT*BIN_COUNT*BIN_COUNT];
	// walk through all descriptors
	char path[MAX_PATH];
	const char *catdir, *catname;
	FILE *file = NULL;
	for (int c = 1; c <= 3; c++)
	{
		// determine the category directory to use
		switch (c)
		{
		case 1: catdir = cat1dir; catname = CATEGORY_1; break;
		case 2: catdir = cat2dir; catname = CATEGORY_2; break;
		case 3: catdir = cat3dir; catname = CATEGORY_3; break;
		}
		cout << "processing category " << catname << endl;
		// average all descriptors
		memset(average, 0, BIN_COUNT*BIN_COUNT*BIN_COUNT * sizeof(float));
		for (int i = 1; i <= 50; i++)
		{
			cout << "processing image " << i << endl;
			// load the histogram descriptor
			SAFE_SPRINTF(path, sizeof(path), "%s%i.dat", catdir, i);
			if ((file = fopen(path, "rb")) == NULL)
				goto failure;
			if (fread(histogram, sizeof(float), BIN_COUNT*BIN_COUNT*BIN_COUNT, file) != BIN_COUNT*BIN_COUNT*BIN_COUNT)
				goto failure;
			SAFE_CLOSEFILE(file);
			// add the value of each bin to the average
			for (int b = 0; b < BIN_COUNT*BIN_COUNT*BIN_COUNT; b++)
				average[b] += histogram[b];
		}
		for (int b = 0; b < BIN_COUNT*BIN_COUNT*BIN_COUNT; b++)
			average[b] /= 50;
		// save the average to disk
		SAFE_SPRINTF(path, sizeof(path), "%s%s.dat", catdir, "average");
		if ((file = fopen(path, "wb")) == NULL)
			goto failure;
		if (fwrite(average, sizeof(float), BIN_COUNT*BIN_COUNT*BIN_COUNT, file) != BIN_COUNT*BIN_COUNT*BIN_COUNT)
			goto failure;
		SAFE_CLOSEFILE(file);
	}
	// release resources
	SAFE_DELETE_ARRAY(histogram);
	SAFE_DELETE_ARRAY(average);
	return true;

failure:
	SAFE_CLOSEFILE(file);
	SAFE_DELETE_ARRAY(histogram);
	SAFE_DELETE_ARRAY(average);
	return false;
}

bool DetermineWinterSport(const char *cat1dir, const char *cat2dir, const char *cat3dir)
{
	float *histogram = new float[BIN_COUNT*BIN_COUNT*BIN_COUNT];
	float *average = new float[BIN_COUNT*BIN_COUNT*BIN_COUNT];
	char path[MAX_PATH] = {0};
	const char *catdir, *catname;
	FILE *file = NULL;
	int c;
	float cat1dist, cat2dist, cat3dist;
	float *catdist;
	// ask for an input image and determine the most likely
	// category it belongs to
	OPENFILENAME ofn = {0};
	//ofn.lpstrFilter = "Image files\0*.jpg;*.png;*.bmp\0\0";
	ofn.lpstrFilter = "Image files\0*.jpg\0\0";
	ofn.lpstrFile = path;
	ofn.nMaxFile = MAX_PATH;
	ofn.lpstrTitle = "Choose image file";
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
	ofn.lStructSize = sizeof(OPENFILENAME);
	if (GetOpenFileName(&ofn) == FALSE)
		goto failure;
	// calculate the histogram of the image
	if (!CalculateDescriptor(path, histogram))
		goto failure;
	// determine the distance to each category
	for (c = 1; c <= 3; c++)
	{
		// determine the category directory to use
		switch (c)
		{
		case 1: catdir = cat1dir; catdist = &cat1dist; break;
		case 2: catdir = cat2dir; catdist = &cat2dist; break;
		case 3: catdir = cat3dir; catdist = &cat3dist; break;
		}
		// load the average from disk
		SAFE_SPRINTF(path, sizeof(path), "%s%s.dat", catdir, "average");
		if ((file = fopen(path, "rb")) == NULL)
			goto failure;
		if (fread(average, sizeof(float), BIN_COUNT*BIN_COUNT*BIN_COUNT, file) != BIN_COUNT*BIN_COUNT*BIN_COUNT)
			goto failure;
		SAFE_CLOSEFILE(file);
		// determine distance
		*catdist = 0.0f;
		for (int b = 0; b < BIN_COUNT*BIN_COUNT*BIN_COUNT; b++)
			*catdist += fabs(histogram[b] - average[b]);
	}
	// determine the winning category
	if (cat1dist < cat2dist)
	{
		if (cat1dist < cat3dist)
			catname = CATEGORY_1;
		else
			catname = CATEGORY_3;
	}
	else if (cat2dist < cat3dist)
		catname = CATEGORY_2;
	else
		catname = CATEGORY_3;
	// print out the results
	cout << "Distance to " << CATEGORY_1 << ": " << cat1dist << endl;
	cout << "Distance to " << CATEGORY_2 << ": " << cat2dist << endl;
	cout << "Distance to " << CATEGORY_3 << ": " << cat3dist << endl;
	cout << "The category this image belongs is category: " << catname << endl;
	cout << "Press any key to continue... ";
	_getch();
	fflush(stdin);
	// release resources
	SAFE_DELETE_ARRAY(histogram);
	SAFE_DELETE_ARRAY(average);
	return true;

failure:
	SAFE_CLOSEFILE(file);
	SAFE_DELETE_ARRAY(histogram);
	SAFE_DELETE_ARRAY(average);
	return false;
}